import os
import sys
from typing import Generator

import pytest
from click.testing import CliRunner as __CliRunner
from hatch.config.constants import AppEnvVars, ConfigEnvVars
from hatch.config.user import ConfigFile
from hatch.project.core import Project
from hatch.utils.fs import Path, temp_directory
from hatch.utils.platform import Platform
from hatch.utils.structures import EnvVars

from .utils import update_project_environment

PLATFORM = Platform()


class CliRunner(__CliRunner):
    def __init__(self, command):
        super().__init__()
        self._command = command

    def __call__(self, *args, **kwargs):
        # Exceptions should always be handled
        kwargs.setdefault('catch_exceptions', False)

        return self.invoke(self._command, args, **kwargs)


@pytest.fixture(scope='session')
def hatch():
    from hatch import cli

    return CliRunner(cli.hatch)


@pytest.fixture(scope='session', autouse=True)
def isolation() -> Generator[Path, None, None]:
    with temp_directory() as d:
        data_dir = d / 'data'
        data_dir.mkdir()
        cache_dir = d / 'cache'
        cache_dir.mkdir()

        default_env_vars = {
            AppEnvVars.NO_COLOR: '1',
            ConfigEnvVars.DATA: str(data_dir),
            ConfigEnvVars.CACHE: str(cache_dir),
            'COLUMNS': '80',
            'LINES': '24',
        }
        with d.as_cwd(default_env_vars):
            os.environ.pop(AppEnvVars.ENV_ACTIVE, None)
            yield d


@pytest.fixture(scope='session')
def data_dir() -> Generator[Path, None, None]:
    yield Path(os.environ[ConfigEnvVars.DATA])


@pytest.fixture(scope='session')
def platform():
    return PLATFORM


@pytest.fixture
def temp_dir() -> Generator[Path, None, None]:
    with temp_directory() as d:
        yield d


@pytest.fixture
def temp_dir_data(temp_dir) -> Generator[Path, None, None]:
    data_path = temp_dir / 'data'
    data_path.mkdir()

    with EnvVars({ConfigEnvVars.DATA: str(data_path)}):
        yield temp_dir


@pytest.fixture(autouse=True)
def config_file(tmp_path) -> ConfigFile:
    path = Path(tmp_path, 'config.toml')
    os.environ[ConfigEnvVars.CONFIG] = str(path)
    config = ConfigFile(path)
    config.restore()
    return config


@pytest.fixture(scope='session')
def default_python_version():
    return f'{sys.version_info[0]}.{sys.version_info[1]}'


@pytest.fixture(scope='session')
def project_name():
    return os.urandom(12).hex()


@pytest.fixture
def conda_project(hatch, temp_dir_data, config_file, project_name):
    config_file.model.template.plugins['default']['tests'] = False
    config_file.save()

    with temp_dir_data.as_cwd():
        result = hatch('new', project_name)

    assert result.exit_code == 0, result.output

    project_path = temp_dir_data / project_name

    project = Project(project_path)
    config = dict(project.config.envs['default'])
    config.update({'type': 'conda'})
    update_project_environment(project, 'default', config)

    yield project_path



from hatch.project.core import Project

from hatch_conda.plugin import CondaEnvironment


class TestPythonVersion:
    def test_default(self, isolation, data_dir, platform, default_python_version):
        env_config = {}
        project = Project(
            isolation,
            config={
                'project': {'name': 'my_app', 'version': '0.0.1'},
                'tool': {'hatch': {'envs': {'default': env_config}}},
            },
        )
        environment = CondaEnvironment(
            isolation, project.metadata, 'default', project.config.envs['default'], {}, data_dir, platform, 0
        )

        assert environment.python_version == default_python_version

    def test_long(self, isolation, data_dir, platform):
        env_config = {'python': '3.10'}
        project = Project(
            isolation,
            config={
                'project': {'name': 'my_app', 'version': '0.0.1'},
                'tool': {'hatch': {'envs': {'default': env_config}}},
            },
        )
        environment = CondaEnvironment(
            isolation, project.metadata, 'default', project.config.envs['default'], {}, data_dir, platform, 0
        )

        assert environment.python_version == '3.10'

    def test_short(self, isolation, data_dir, platform):
        env_config = {'python': '310'}
        project = Project(
            isolation,
            config={
                'project': {'name': 'my_app', 'version': '0.0.1'},
                'tool': {'hatch': {'envs': {'default': env_config}}},
            },
        )
        environment = CondaEnvironment(
            isolation, project.metadata, 'default', project.config.envs['default'], {}, data_dir, platform, 0
        )

        assert environment.python_version == '3.10'



from __future__ import annotations

import sys
from contextlib import contextmanager

from hatch.env.plugin.interface import EnvironmentInterface


class CondaEnvironment(EnvironmentInterface):
    PLUGIN_NAME = 'conda'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._config_command = None
        self._config_conda_forge = None
        self.__python_version = None

        self.conda_env_name = f'{self.metadata.core.name}_{self.name}_{self.python_version}'
        self.project_path = '.'

    @staticmethod
    def get_option_types():
        return {'command': str, 'conda-forge': bool}

    def _config_value(self, field_name, default, valid=None):
        class_name = f'_config_{field_name.replace("-", "_")}'
        if self.__dict__[class_name] is None:
            value = self.config.get(field_name, default)
            if not isinstance(value, self.get_option_types()[field_name]):
                raise TypeError(
                    f'Field `tool.hatch.envs.{self.name}.{field_name}` must be a '
                    + '`{self.get_option_types()[field_name]}`'
                )
            if valid is not None and value not in valid:
                raise ValueError(f'Field `tool.hatch.envs.{self.name}.{field_name}` must be any of [{valid}] values.')
            self.__dict__[class_name] = value
        return self.__dict__[class_name]

    @property
    def config_command(self):
        return self._config_value('command', 'conda', ['conda', 'mamba'])

    @property
    def config_conda_forge(self):
        return self._config_value('conda-forge', True)

    @property
    def python_version(self):
        if self.__python_version is None:
            python_version = self.config.get('python', '')
            if not python_version:
                python_version = '.'.join(map(str, sys.version_info[:2]))
            elif python_version.isdigit() and len(python_version) > 1:
                python_version = f'{python_version[0]}.{python_version[1:]}'

            self.__python_version = python_version

        return self.__python_version

    def _get_conda_env_path(self, name):
        output = self.platform.check_command_output([self.config_command, 'env', 'list'])
        env_names, env_paths = zip(
            *[(line.split(' ')[0], line.split(' ')[-1]) for line in output.splitlines() if len(line.split(' ')[0]) > 1]
        )
        if name not in env_names:
            return None
        return env_paths[env_names.index(name)]

    def find(self):
        return self._get_conda_env_path(self.conda_env_name)

    def create(self):
        command = [
            self.config_command,
            'create',
            '-y',
            '-n',
            self.conda_env_name,
        ]
        if self.config_conda_forge:
            command += [
                '-c',
                'conda-forge',
                '--no-channel-priority',
            ]
        command += [
            f'python={self.python_version}',
            'pip',
        ]
        if self.verbosity > 0:  # no cov
            self.platform.check_command(command)
        else:
            self.platform.check_command_output(command)

    def remove(self):
        self.platform.check_command_output([self.config_command, 'env', 'remove', '-y', '--name', self.conda_env_name])

    def exists(self):
        return bool(self._get_conda_env_path(self.conda_env_name))

    def construct_conda_run_command(self, command):
        return [self.config_command, 'run', '-n', self.conda_env_name] + command

    def construct_pip_install_command(self, *args, **kwargs):
        return self.construct_conda_run_command(super().construct_pip_install_command(*args, **kwargs))

    def install_project(self):
        with self:
            self.platform.check_command(self.construct_pip_install_command([self.apply_features(self.project_path)]))

    def install_project_dev_mode(self):
        with self:
            self.platform.check_command(
                self.construct_pip_install_command(['--editable', self.apply_features(self.project_path)])
            )

    def dependencies_in_sync(self):
        if not self.dependencies:
            return True

        with self:
            process = self.platform.run_command(
                ' '.join(['hatchling', 'dep', 'synced', '-p', 'python', *self.dependencies]),
                capture_output=True,
            )
            return not process.returncode

    def sync_dependencies(self):
        with self:
            self.platform.check_command(self.construct_pip_install_command(self.dependencies))

    @contextmanager
    def command_context(self):
        with self:
            yield

    def run_shell_command(self, command):
        return self.platform.run_command(
            ' '.join(
                self.construct_conda_run_command(
                    [
                        command,
                    ]
                )
            )
        )

    def enter_shell(self, name, path, args):  # no cov
        with self:
            process = self.platform.run_command(' '.join([self.config_command, 'activate', self.conda_env_name]))
            self.platform.exit_with_code(process.returncode)
